#!/usr/local/bin/python
# encoding: utf-8
"""
*Listen to GCN LV Alerts*

:Author:
    David Young

:Date Created:
    March  3, 2019
"""
################# GLOBAL IMPORTS ####################
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
import gcn
from ENGRAVEAlerts import sms
import lxml.etree
from datetime import datetime
import requests
import operator
# THE LVC-GCN HANDLER
try:
    # for Python 2.x
    from StringIO import StringIO
except ImportError:
    # for Python 3.x
    from io import StringIO
import unicodecsv as csv
import bitly_api


class gcnListener():
    """
    *The worker class for the gcnListener module*

    **Key Arguments:**
        - ``log`` -- logger
        - ``settings`` -- the settings dictionary
        - ``test`` -- use test settings for development purposes

    **Usage:**

        To setup your logger, settings and database connections, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_).

        To initiate a gcnListener object, use the following:

        .. code-block:: python

            from ENGRAVEAlerts import gcnListener
            alerts = gcnListener(
                log=log,
                settings=settings
            )
    """

    def __init__(
            self,
            log,
            settings=False,
            test=False
    ):
        # OBJECT VARIABLES
        self.log = log
        log.debug("instansiating a new 'gcnListener' object")
        self.settings = settings
        self.eventCache = []
        self.contacts = []
        self.test = test

        # INITIAL ACTIONS - CACHE THE CONTACTS FROM GOOGLE SHEET (SHEET ID IN
        # SETTING FILE)
        self.cache_alert_contacts()

        return None

    def listen(
            self,
            testEventPacket=False):
        """*listen to the GCN notice stream and trigger SMS alerts if new GW event discovered*

        **Key Arguments:**
            - ``testEventPacket`` -- use a local VOEvent Packet to test the process_gcn function

        **Return:**
            - None

        **Usage:**

            .. code-block:: python

                from ENGRAVEAlerts import gcnListener
                alerts = gcnListener(
                    log=log,
                    settings=settings
                )
                alerts.listen()
        """
        self.log.debug('starting the ``listen`` method')

        # FILTER THE GCNs TO ONLY LVC ORIGIN
        @gcn.handlers.include_notice_types(
            gcn.notice_types.LVC_PRELIMINARY,
            gcn.notice_types.LVC_INITIAL)
        def process_gcn(payload, root):

            smsContent = ""

            # DECIDE HOW TO RESPOND FOR REAL/TEST EVENTS
            if root.attrib['role'] == 'observation':
                pass
            if root.attrib['role'] == 'test':
                # IF THIS IS NOT A TEST THEN IGNORE THE GCN TEST STREAM
                if not self.test:
                    return
                smsContent = "TEST ALERT: "

            # READ ALL OF THE VOEVENT PARAMETERS FROM THE "WHAT" SECTION.
            params = {elem.attrib['name']:
                      elem.attrib['value']
                      for elem in root.iterfind('.//Param')}

            try:
                # TRY AND PARSE ORDERED CLASSIFICATIONS
                classList = ["BNS", "NSBH", "BBH", "MassGap", "Terrestrial"]
                classifications = {}

                for c in classList:
                    classifications[c] = float(params[c])

                sorted_classifications = sorted(
                    classifications.items(), key=operator.itemgetter(1), reverse=True)

                classifictions = "\n\nClassification of signal in desc. order of prob is "
                for i, v in enumerate(sorted_classifications):
                    k = v[0]
                    v = float(v[1]) * 100
                    if float(v) < 1:
                        v = "<1"
                    else:
                        v = int(round(v))
                    if i == 4:
                        classifictions += " or "
                    elif i > 0:
                        classifictions += ", "
                    classifictions += "%s (%s%%)" % (k, v)
            except:
                pass

            try:
                # CONVERT FAR TO YEARS
                FAR = float(params["FAR"])
                FAR = 1. / (FAR * 60 * 60 * 24 * 365)

                if FAR < 10:
                    FAR = "%(FAR)1.2f" % locals()
                elif FAR < 100:
                    FAR = "%(FAR)1.1f" % locals()
                else:
                    FAR = int(round(FAR))
                params["FAR"] = FAR
            except:
                pass

            # PARSE THE EVENT TIME
            eventTime = "?"
            for elem in root.iterfind('.//ISOTime'):
                eventTime = elem.text
                try:
                    eventTime = datetime.strptime(
                        eventTime, '%Y-%m-%dT%H:%M:%S.%f')
                    eventTime = eventTime.strftime('%Y-%m-%d %H:%M:%S')
                except:
                    print "Could not convert time stamp"
                    pass
                eventTime = eventTime + " UT"
            params["eventTime"] = eventTime

            # BUILD ALERT CONTENT
            smsContent += "%(AlertType)s notice for %(GraceID)s, a %(Group)s event detected at %(eventTime)s.\n\nFAR=1 every %(FAR)s yrs." % params

            if params["Group"] == "CBC":
                smsContent += classifictions

            # SLACK URL
            # smsContent += "https://bit.ly/2VZdoWL  "

            API_USER = self.settings["bitly"]["user"]
            API_KEY = self.settings["bitly"]["api_key"]
            try:
                b = bitly_api.Connection(API_USER, API_KEY)
                # SHORTEN GRACEDB URL
                longurl = params["EventPage"]
                params["EventPage"] = b.shorten(uri=longurl)["url"]
            except:
                pass

            try:
                smsContent += " %(EventPage)s" % params
            except:
                pass

            # PRINT ALL PARAMETERS TO STDOUT
            for key, value in params.items():
                print key, ':', value

            # ONLY TRIGGER SMS ON NEW EVENTS
            if params["GraceID"] not in self.eventCache:
                # TRY A REPARSE OF THE ALERT CONTACTS - USE CACHE IF FAIL
                try:
                    self.cache_alert_contacts()
                except:
                    pass

                # SEND SMS ALERTS
                alerts = sms(
                    log=self.log,
                    settings=self.settings
                )
                alerts.set_mobile_numbers(self.contacts)
                alerts.send_message(smsContent)

            # CACHE THE EVENT ID SO NOT TO DUPLICATE ALERTS
            self.eventCache.append(params["GraceID"])

        # TRIGGER process_gcn ONCE IF TEST ALERT PACKET GIVEN - ELSE OPEN GCN
        # LISTENER
        if not testEventPacket:
            gcn.listen(handler=process_gcn)
        else:
            payload = open(testEventPacket, 'rb').read()
            root = lxml.etree.fromstring(payload)
            process_gcn(payload, root)

        self.log.debug('completed the ``listen`` method')
        return None

    def cache_alert_contacts(
            self):
        """*Download a copy of the alert contacts from google sheet and parse numbers into an array*

        **Return:**
            - None

        **Usage:**

            .. code-block:: python 

                from ENGRAVEAlerts import gcnListener
                alerts = gcnListener(
                    log=log,
                    settings=settings
                )
                alerts.cache_alert_contacts()
        """
        self.log.debug('starting the ``cache_alert_contacts`` method')

        # GET THE RELEVANT GOOGLE SHEET ID (DEV OR PRODUCTION VERSION)
        if self.test:
            url = self.settings["contact_sheet"]["development"]
        else:
            url = self.settings["contact_sheet"]["production"]

        # REQUEST THE CONTENT OF THE SHEET
        content = False
        try:
            response = requests.get(
                url=url,
            )
            content = response.content
            status_code = response.status_code
        except requests.exceptions.RequestException:
            print 'HTTP Request failed'

        # SOMETIMES THE CONTENT HEADER ISN'T RETURNED CORRECTLY
        content = content.replace(
            '"Name","","",', '"Name","Mobile (with country code)","SMS Alerts",')

        if content:
            # READ AS CSV CONTENT
            f = StringIO(content)
            csvReader = csv.DictReader(
                f, dialect='excel', delimiter=',', quotechar='"')
            self.contacts = []
            for row in csvReader:
                # IF SMS TOGGLE IS ON THEN NUMBER APPEND TO ARRAY
                if "SMS Alerts" in row and row["SMS Alerts"] == "TRUE":
                    if "+" not in row["Mobile (with country code)"]:
                        row["Mobile (with country code)"] = "+" + \
                            row["Mobile (with country code)"]
                    self.contacts.append(row["Mobile (with country code)"])

        self.log.debug('completed the ``cache_alert_contacts`` method')
        return None

    # use the tab-trigger below for new method
    # xt-class-method



#!/usr/local/bin/python
# encoding: utf-8
"""
*Send SMS Alerts to ENGRAVErs*

:Author:
    David Young

:Date Created:
    March  1, 2019
"""
################# GLOBAL IMPORTS ####################
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from twilio.rest import Client
import time


class sms():
    """
    *The worker class for the sms module*

    **Key Arguments:**
        - ``log`` -- logger
        - ``settings`` -- the settings dictionary

    **Usage:**

        To setup your logger, settings and database connections, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_). 

        To initiate a sms object, use the following:

        .. todo::

            - add a tutorial about ``sms`` to documentation

        .. code-block:: python 

            from ENGRAVEAlerts import sms
            alerts = sms(
                log=log,
                settings=settings
            )
    """
    # Initialisation

    def __init__(
            self,
            log,
            settings=False
    ):
        self.log = log
        log.debug("instansiating a new 'sms' object")
        self.settings = settings
        self.numbers = []
        self.mynumber = self.settings["twilio"][
            "phone"]
        self.mynumber2 = self.settings["twilio"][
            "phone2"]
        self.myid = self.settings["twilio"][
            "smsid"]
        self.client = Client(self.settings["twilio"][
                             "sid"], self.settings["twilio"]["auth_token"])

        return None

    def set_mobile_numbers(
            self,
            numbers):
        """*set the recipients of the SMS alert as either a single number or a list of numbers*

        **Key Arguments:**
            - ``numbers`` -- either a single mobile number as a string, a comma separated string or a list of numbers.

        **Usage:**

            .. code-block:: python 

                from ENGRAVEAlerts import sms
                this = sms(
                    log=log,
                    settings=settings
                )
                this.set_mobile_numbers("+4479134471, +44791253271, +44791345391")
        """
        self.log.debug('starting the ``set_mobile_numbers`` method')

        if isinstance(numbers, str):
            numbers = numbers.replace(",", " ")
            numbers = numbers.split()

        tmp = []
        tmp[:] = [t.strip() for t in numbers]
        self.numbers = tmp

        self.log.debug('completed the ``set_mobile_numbers`` method')
        return None

    def send_message(
            self,
            message):
        """*send a message (160 characters) to the list of mobile numbers*

        **Key Arguments:**
            - ``message`` -- the message (will be clipped to 160 characters)

        **Usage:**

            .. code-block:: python 

                from ENGRAVEAlerts import sms
                alerts = sms(
                    log=log,
                    settings=settings
                )
                alerts.set_mobile_numbers("+4479134471, +44791253271, +44791345391")
                alerts.send_message("this is a test from the ENGRAVEAlerts package")
        """
        self.log.debug('starting the ``send_message`` method')

        if len(self.numbers) == 0:
            log.error('cound not send sms - no mobile numbers were given')
            return None

        for n in self.numbers:
            sent = False
            try:
                sms = self.client.messages.create(
                    body=message[0:300],
                    from_=self.mynumber,
                    to=n
                )
                sent = True
            except Exception, e:
                print "Can't send to %(n)s" % locals()
                print "%(e)s" % locals()
                pass

            if not sent:
                # TRY THE BACKUP NUMBER IF DEFAULT NUMBER DIDN'T WORK (e.g.
                # WITH NORTH AMERICAN NUMBERS)
                try:
                    sms = self.client.messages.create(
                        body=message[0:300],
                        from_=self.mynumber2,
                        to=n
                    )
                    sent = True
                except Exception, e:
                    print "Can't send to %(n)s" % locals()
                    print "%(e)s" % locals()
                    pass

        self.log.debug('completed the ``send_message`` method')
        return None



import os
import nose2
import shutil
import unittest
import yaml
from ENGRAVEAlerts import gcnListener, cl_utils
from ENGRAVEAlerts.utKit import utKit

from fundamentals import tools

su = tools(
    arguments={"settingsFile": None},
    docString=__doc__,
    logLevel="DEBUG",
    options_first=False,
    projectName="ENGRAVEAlerts",
    defaultSettingsFile=False
)
arguments, settings, log, dbConn = su.setup()

# # load settings
# stream = file(
#     "/Users/Dave/.config/ENGRAVEAlerts/ENGRAVEAlerts.yaml", 'r')
# settings = yaml.load(stream)
# stream.close()

# SETUP AND TEARDOWN FIXTURE FUNCTIONS FOR THE ENTIRE MODULE
moduleDirectory = os.path.dirname(__file__)
utKit = utKit(moduleDirectory)
log, dbConn, pathToInputDir, pathToOutputDir = utKit.setupModule()
utKit.tearDownModule()

# load settings
stream = file(
    "/Users/Dave/Dropbox/config/dave-macbook/ENGRAVEAlerts/ENGRAVEAlerts.yaml", 'r')
settings = yaml.load(stream)
stream.close()

import shutil
try:
    shutil.rmtree(pathToOutputDir)
except:
    pass
# COPY INPUT TO OUTPUT DIR
shutil.copytree(pathToInputDir, pathToOutputDir)

# Recursively create missing directories
if not os.path.exists(pathToOutputDir):
    os.makedirs(pathToOutputDir)

# xt-setup-unit-testing-files-and-folders


class test_gcnListener(unittest.TestCase):

    def test_gcnListener_function(self):

        from ENGRAVEAlerts import gcnListener
        this = gcnListener(
            log=log,
            settings=settings,
            test=True
        )
        this.get()

    def test_listen_function(self):

        from ENGRAVEAlerts import gcnListener
        alerts = gcnListener(
            log=log,
            settings=settings
        )
        alerts.listen()

    def test_process_gcn_function(self):

        testEventPacket = pathToInputDir + '/Preliminary-VOEvent.xml'
        from ENGRAVEAlerts import gcnListener
        alerts = gcnListener(
            log=log,
            settings=settings,
            test=True
        )
        alerts.listen(testEventPacket=testEventPacket)

    def test_cache_contacts_function(self):

        from ENGRAVEAlerts import gcnListener
        alerts = gcnListener(
            log=log,
            settings=settings,
            test=True
        )
        alerts.cache_alert_contacts()

    def test_process_gcn_function2(self):

        testEventPacket = pathToInputDir + '/Preliminary-VOEvent.xml'
        from ENGRAVEAlerts import gcnListener
        alerts = gcnListener(
            log=log,
            settings=settings,
            test=True
        )
        alerts.listen(testEventPacket=testEventPacket)

        testEventPacket = pathToInputDir + '/Initial-VOEvent.xml'
        alerts.listen(testEventPacket=testEventPacket)

        testEventPacket = pathToInputDir + '/Update-VOEvent.xml'
        alerts.listen(testEventPacket=testEventPacket)

        testEventPacket = pathToInputDir + '/Retraction-VOEvent.xml'
        alerts.listen(testEventPacket=testEventPacket)

    def test_gcnListener_function_exception(self):

        from ENGRAVEAlerts import gcnListener
        try:
            this = gcnListener(
                log=log,
                settings=settings,
                fakeKey="break the code"
            )
            this.get()
            assert False
        except Exception, e:
            assert True
            print str(e)

        # x-print-testpage-for-pessto-marshall-web-object

    # x-class-to-test-named-worker-function



#!/usr/local/bin/python
# encoding: utf-8
"""
*Parse the GCN notice and convert into a human readable format*

:Author:
    David Young

:Date Created:
    March 21, 2019
"""
################# GLOBAL IMPORTS ####################
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools


# OR YOU CAN REMOVE THE CLASS BELOW AND ADD A WORKER FUNCTION ... SNIPPET TRIGGER BELOW
# xt-worker-def

class gcnNoticeParser():
    """
    *The worker class for the gcnNoticeParser module*

    **Key Arguments:**
        - ``log`` -- logger
        - ``settings`` -- the settings dictionary

    **Usage:**

        To setup your logger, settings and database connections, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_). 

        To initiate a gcnNoticeParser object, use the following:

        .. todo::

            - add usage info
            - create a sublime snippet for usage
            - create cl-util for this class
            - add a tutorial about ``gcnNoticeParser`` to documentation
            - create a blog post about what ``gcnNoticeParser`` does

        .. code-block:: python 

            usage code   
    """
    # Initialisation
    # 1. @flagged: what are the unique attrributes for each object? Add them
    # to __init__

    def __init__(
            self,
            log,
            settings=False,

    ):
        self.log = log
        log.debug("instansiating a new 'gcnNoticeParser' object")
        self.settings = settings
        # xt-self-arg-tmpx

        # 2. @flagged: what are the default attrributes each object could have? Add them to variable attribute set here
        # Variable Data Atrributes

        # 3. @flagged: what variable attrributes need overriden in any baseclass(es) used
        # Override Variable Data Atrributes

        # Initial Actions

        return None

    # 4. @flagged: what actions does each object have to be able to perform? Add them here
    # Method Attributes
    def get(self):
        """
        *get the gcnNoticeParser object*

        **Return:**
            - ``gcnNoticeParser``

        **Usage:**
        .. todo::

            - add usage info
            - create a sublime snippet for usage
            - create cl-util for this method
            - update the package tutorial if needed

        .. code-block:: python 

            usage code 
        """
        self.log.debug('starting the ``get`` method')

        gcnNoticeParser = None

        self.log.debug('completed the ``get`` method')
        return gcnNoticeParser

    # xt-class-method

    # 5. @flagged: what actions of the base class(es) need ammending? ammend them here
    # Override Method Attributes
    # method-override-tmpx



#!/usr/local/bin/python
# encoding: utf-8
"""
*Send voicecall alert to ENGRAVErs*

:Author:
    David Young

:Date Created:
    March  1, 2019
"""
################# GLOBAL IMPORTS ####################
import sys
import os
os.environ['TERM'] = 'vt100'
from fundamentals import tools
from ENGRAVEAlerts import sms
import urllib
import twilio

# OR YOU CAN REMOVE THE CLASS BELOW AND ADD A WORKER FUNCTION ... SNIPPET TRIGGER BELOW
# xt-worker-def


class voicecall(sms.sms):
    """
    *The worker class for the voicecall module*

    **Key Arguments:**
        - ``log`` -- logger
        - ``settings`` -- the settings dictionary

    **Usage:**

        To setup your logger, settings and database connections, please use the ``fundamentals`` package (`see tutorial here <http://fundamentals.readthedocs.io/en/latest/#tutorial>`_). 

        To initiate a voicecall object, use the following:

        .. todo::

            - add a tutorial about ``voicecall`` to documentation

        .. code-block:: python 

            from ENGRAVEAlerts import voicecall
            alerts = voicecall(
                log=log,
                settings=settings
            )
    """

    def make_voicecall(
            self,
            message):
        """*make a voice call with a short message*

        **Key Arguments:**
            - ``message`` -- the message to speak is recipient answers

        **Return:**
            - None

        **Usage:**

            .. code-block:: python 

                from ENGRAVEAlerts import voicecall
                alerts = voicecall(
                    log=log,
                    settings=settings
                )
                alerts.set_mobile_numbers(
                    "+44795432071, +4474355071, +44343553371")
                alerts.make_voicecall(
                    "this is a test from the ENGRAVEAlerts package")

        """
        self.log.debug('starting the ``make_voicecall`` method')

        if len(self.numbers) == 0:
            log.error('cound not make voice call - no mobile numbers were given')
            return None
        for n in self.numbers:

            try:
                messageForCall = urllib.quote(message) + "&"
                call = self.client.calls.create(to=n,
                                                from_=self.mynumber,
                                                url="http://twimlets.com/message?Message%5B0%5D=" + messageForCall)
            except Exception, e:
                print "Can't call %(n)s" % locals()
                print "%(e)s" % locals()
                pass

        self.log.debug('completed the ``make_voicecall`` method')
        return None



#!/usr/local/bin/python
# encoding: utf-8
"""
Documentation for ENGRAVEAlerts can be found here: http://ENGRAVEAlerts.readthedocs.org/en/stable

Usage:
    ealert init
    ealert sms <mobilenumbers> <message> [-s <pathToSettingsFile>]
    ealert call <mobilenumbers> <message> [-s <pathToSettingsFile>]
    ealert listen [-t] [-s <pathToSettingsFile>]

Options:
    init                  setup the ENGRAVEAlerts settings file for the first time
    sms                   send sms
    listen                listen for and alert on new events

    <mobilenumbers>       quoted mobile numbers space or comma separated
    <message>             quoted message (160 limit)

    -t, --test            run in test mode
    -h, --help            show this help message
    -v, --version         show version
    -s, --settings        the settings file
"""
################# GLOBAL IMPORTS ####################
import sys
import os
os.environ['TERM'] = 'vt100'
import readline
import glob
import pickle
from docopt import docopt
from fundamentals import tools, times
from subprocess import Popen, PIPE, STDOUT
# from ..__init__ import *
from ENGRAVEAlerts.sms import sms as esms
from ENGRAVEAlerts.voicecall import voicecall


def tab_complete(text, state):
    return (glob.glob(text + '*') + [None])[state]


def main(arguments=None):
    """
    *The main function used when ``cl_utils.py`` is run as a single script from the cl, or when installed as a cl command*
    """
    # SETUP THE COMMAND-LINE UTIL SETTINGS
    su = tools(
        arguments=arguments,
        docString=__doc__,
        logLevel="DEBUG",
        options_first=False,
        projectName="ENGRAVEAlerts",
        defaultSettingsFile=True
    )
    arguments, settings, log, dbConn = su.setup()

    # UNPACK REMAINING CL ARGUMENTS USING `EXEC` TO SETUP THE VARIABLE NAMES
    # AUTOMATICALLY
    for arg, val in arguments.iteritems():
        if arg[0] == "-":
            varname = arg.replace("-", "") + "Flag"
        else:
            varname = arg.replace("<", "").replace(">", "")
        if varname == "import":
            varname = "iimport"
        if isinstance(val, str) or isinstance(val, unicode):
            exec(varname + " = '%s'" % (val,))
        else:
            exec(varname + " = %s" % (val,))
        if arg == "--dbConn":
            dbConn = val
        log.debug('%s = %s' % (varname, val,))

    ## START LOGGING ##
    startTime = times.get_now_sql_datetime()
    log.info(
        '--- STARTING TO RUN THE cl_utils.py AT %s' %
        (startTime,))

    if init:
        from os.path import expanduser
        home = expanduser("~")
        filepath = home + "/.config/ENGRAVEAlerts/ENGRAVEAlerts.yaml"
        try:
            cmd = """open %(filepath)s""" % locals()
            p = Popen(cmd, stdout=PIPE, stderr=PIPE, shell=True)
        except:
            pass
        try:
            cmd = """start %(filepath)s""" % locals()
            p = Popen(cmd, stdout=PIPE, stderr=PIPE, shell=True)
        except:
            pass
        return

    # SEND AN SMS TO THE NUMBERS REQUESTED
    if sms:
        alerts = esms(
            log=log,
            settings=settings
        )
        alerts.set_mobile_numbers(mobilenumbers)
        alerts.send_message(message)

    # CALL THE NUMBERS REQUESTED
    if call:
        from ENGRAVEAlerts import voicecall
        alerts = voicecall(
            log=log,
            settings=settings
        )
        alerts.set_mobile_numbers(mobilenumbers)
        alerts.make_voicecall(message)

    # LISTEN TO THE GCN NOTICES FOR LV AND SEND AN SMS IF TRIGGERED
    if listen:
        from ENGRAVEAlerts import gcnListener
        alerts = gcnListener(
            log=log,
            settings=settings,
            test=testFlag
        )
        alerts.listen()

    # CALL FUNCTIONS/OBJECTS

    if "dbConn" in locals() and dbConn:
        dbConn.commit()
        dbConn.close()
    ## FINISH LOGGING ##
    endTime = times.get_now_sql_datetime()
    runningTime = times.calculate_time_difference(startTime, endTime)
    log.info('-- FINISHED ATTEMPT TO RUN THE cl_utils.py AT %s (RUNTIME: %s) --' %
             (endTime, runningTime, ))

    return


if __name__ == '__main__':
    main()
